@@ -52,10 +52,15 @@ bool edspLikeListParser::NewVersion(pkgCache::VerIterator &Ver)
       if (version != Ver.VerStr())
       {
 	 map_stringitem_t const idx = StoreString(pkgCacheGenerator::VERSIONNUMBER, version);
-	 Ver->SourceVerStr = idx;
+	 Ver.SourceVersion()->VerStr = idx;
       }
    }
 
+   // If we have a Size field, use it, otherwise fake one based on APT-Release to
+   // be able to distinguish downloadable debs from installed ones.
+   Ver->Size = Section.Exists(pkgTagSection::Key::Size)
+		  ? Section.FindI(pkgTagSection::Key::Size)
+		  : Section.Exists("APT-Release");
    return true;
 }
 									/*}}}*/
