@@ -35,6 +35,7 @@
 #include <cassert>
 #include <chrono>
 #include <ctime>
+#include <iomanip>
 #include <sstream>
 
 // FIXME: Helpers stolen from DepCache, please give them back.
@@ -134,7 +135,7 @@ struct APT::Solver::CompareProviders3 /*{{{*/
       if ((A->CurrentVer == 0 || B->CurrentVer == 0) && A->CurrentVer != B->CurrentVer)
 	 return A->CurrentVer != 0;
       // Prefer packages in the same group as the target; e.g. foo:i386, foo:amd64
-      if (A->Group != B->Group)
+      if (A->Group != B->Group && not Pkg.end())
       {
 	 if (A->Group == Pkg->Group && B->Group != Pkg->Group)
 	    return true;
@@ -215,13 +216,15 @@ APT::Solver::Solver(pkgCache &cache, pkgDepCache::Policy &policy, EDSP::Request:
    rootState->decision = Decision::MUST;
 }
 
+APT::Solver::~Solver() = default;
+
 // This function determines if a work item is less important than another.
 bool APT::Solver::Work::operator<(APT::Solver::Work const &b) const
 {
    if ((not clause->optional && size < 2) != (not b.clause->optional && b.size < 2))
       return not b.clause->optional && b.size < 2;
-   if (clause->optional != b.clause->optional)
-      return clause->optional;
+   if (clause->eager != b.clause->eager)
+      return not clause->eager;
    if (clause->group != b.clause->group)
       return clause->group > b.clause->group;
    if ((size < 2) != (b.size < 2))
@@ -231,16 +234,49 @@ bool APT::Solver::Work::operator<(APT::Solver::Work const &b) const
    return false;
 }
 
-std::string APT::Solver::Clause::toString(pkgCache &cache) const
+std::string APT::Solver::Clause::toString(pkgCache &cache, bool pretty, bool showMerged) const
 {
    std::string out;
-   if (auto Pkg = reason.Pkg(cache); not Pkg.end())
-      out.append(Pkg.FullName());
-   if (auto Ver = reason.Ver(cache); not Ver.end())
-      out.append(Ver.ParentPkg().FullName()).append("=").append(Ver.VerStr());
-   out.append(" -> ");
-   for (auto var : solutions)
-      out.append(" | ").append(var.toString(cache));
+   if (showMerged)
+      out.append(reason.toString(cache));
+   if (dep && pretty)
+   {
+      out.append(" ").append(pkgCache::DepIterator(cache, dep).DepType()).append(" ");
+      for (auto dep = pkgCache::DepIterator(cache, this->dep); not dep.end(); ++dep)
+      {
+	 out.append(dep.TargetPkg().FullName(true));
+	 if (dep.TargetVer())
+	    out.append(" (").append(dep.CompType()).append(" ").append(dep.TargetVer()).append(")");
+	 if (!(dep->CompareOp & pkgCache::Dep::Or))
+	    break;
+	 out.append(" | ");
+      }
+   }
+   else if (pretty && group == Group::SelectVersion && negative)
+      out.append(" conflicts with other versions of itself");
+   else if (pretty && group == Group::SelectVersion && reason.Pkg()) {
+      out.append(solutions.size() > 1 ? " is available in versions " : " is available in version ");
+      for (auto var : solutions) {
+	 assert(var.Ver());
+	 if (var != solutions.front())
+	    out.append(", ");
+	 out.append(var.Ver(cache).VerStr());
+      }
+   }
+   else
+   {
+      out.append(" -> ");
+      for (auto var : solutions)
+	 out.append(" | ").append(var.toString(cache));
+   }
+   if (showMerged && not merged.empty())
+   {
+      for (auto &clause : merged)
+      {
+	 out.append(" and");
+	 out.append(clause.toString(cache, pretty, false));
+      }
+   }
    return out;
 }
 
@@ -262,8 +298,12 @@ inline APT::Solver::Var APT::Solver::bestReason(APT::Solver::Clause const *claus
       return Var{};
    if (clause->reason == var)
       for (auto choice : clause->solutions)
-	 if ((*this)[choice].decision != Decision::NONE)
+      {
+	 if (clause->negative && (*this)[choice].decision == Decision::MUST)
 	    return choice;
+	 if (not clause->negative && (*this)[choice].decision == Decision::MUSTNOT)
+	    return choice;
+      }
    return clause->reason;
 }
 
@@ -288,6 +328,168 @@ std::string APT::Solver::WhyStr(Var reason) const
    return outstr;
 }
 
+std::string APT::Solver::LongWhyStr(Var var, bool decision, const Clause *rclause, std::string prefix, std::unordered_set<Var> &seen) const
+{
+   std::ostringstream out;
+
+   // Helper function to nicely print more details than just "install/do not install", such as "removal", "upgrade", "downgrade", "install"
+   auto printSelection = [this](Var var, bool decision)
+   {
+      std::string s;
+      if (auto pkg = var.Pkg(cache); not decision && pkg && pkg->CurrentVer)
+	 strprintf(s, "%s is selected for removal", var.toString(cache).c_str());
+      else if (auto ver = var.Ver(cache); decision && ver && ver.ParentPkg().CurrentVer() && ver.ParentPkg().CurrentVer() != ver)
+      {
+	 if (cache.VS->CmpVersion(ver.ParentPkg().CurrentVer().VerStr(), ver.VerStr()) < 0)
+	    strprintf(s, "%s is selected as an upgrade", var.toString(cache).c_str());
+	 else
+	    strprintf(s, "%s is selected as a downgrade", var.toString(cache).c_str());
+      }
+      else if (not decision)
+	 strprintf(s, "%s is not selected for install", var.toString(cache).c_str());
+      else
+	 strprintf(s, "%s is selected for install", var.toString(cache).c_str());
+      return s;
+   };
+
+   // Helper: Recurse into all of the children of the clause and print the decision for them.
+   auto recurseChildren = [&](const Clause *clause, Var skip = Var())
+   {
+      if (clause->solutions.empty())
+	 out << prefix << "[no choices]\n";
+      for (auto choice : clause->solutions)
+      {
+	 if (choice == skip)
+	    continue;
+
+	 if ((*this)[choice].decision == Decision::NONE)
+	    out << prefix << "- " << choice.toString(cache) << " is undecided\n";
+	 else
+	    out << prefix << "- " << LongWhyStr(choice, (*this)[choice].decision == Decision::MUST, (*this)[choice].reason, prefix + "  ", seen).substr(prefix.size() + 2);
+      }
+   };
+
+   // Inverse version selection clauses that select the package if the version is selected,
+   // such as pkg=ver -> pkg, are irrelevant for the user, skip them
+   if (var.Pkg() && decision && rclause && rclause->group == Group::SelectVersion)
+   {
+      var = rclause->reason;
+      rclause = (*this)[var].reason;
+   }
+
+   // No reason given, probably a user request or manually installed or essential or whatnot.
+   if (not rclause)
+   {
+      out << prefix << printSelection(var, decision) << "\n";
+      return out.str();
+   }
+
+   // We could be called with a decision we tried to make but failed due to a conflict;
+   // this checks if it is the real decision.
+   if ((*this)[var].decision != Decision::NONE && decision == ((*this)[var].decision == Decision::MUST) && (*this)[var].reason == rclause)
+   {
+      // If we have seen the real decision before; we dont't need to print it again.
+      if (seen.find(var) != seen.end())
+      {
+	 out << prefix << printSelection(var, decision) << " as above\n";
+	 return out.str();
+      }
+      seen.insert(var);
+   }
+
+   // A package was decided "not install" due to a positive clause, so the clause is unsat.
+   if (not decision && rclause && not rclause->negative)
+   {
+      out << prefix << rclause->toString(cache, true) << "\n";
+      out << prefix << "but none of the choices are installable:\n";
+      recurseChildren(rclause);
+      return out.str();
+   }
+
+   // Build the strongest path from a root to our decision leaf
+   std::vector<Var> path;
+   for (auto reason = bestReason(rclause, var); not reason.empty(); reason = bestReason((*this)[reason].reason, reason))
+      path.push_back(reason);
+
+   // Render the strong reasoning path
+   out << prefix << printSelection(var, decision) << " because:\n";
+   auto w = std::to_string(path.size() + 1).size();
+   size_t i = 1;
+   for (auto it = path.rbegin(); it != path.rend(); ++it, ++i)
+   {
+      auto const &state = (*this)[*it];
+      // Don't print version selection clauses
+      if (it->Pkg() && state.reason && state.reason->group == Group::SelectVersion)
+      {
+	 --i;
+	 continue;
+      }
+      if (seen.find(*it) != seen.end())
+      {
+	 if ((it + 1) == path.rend() || seen.find(*(it + 1)) == seen.end())
+	 {
+	    out << prefix << (i == 1 ? "" : "1-") << i << ". " << printSelection(*it, state.decision == Decision::MUST) << " as above\n";
+	 }
+	 continue;
+      }
+      seen.insert(*it);
+      if (state.reason)
+      {
+	 out << prefix << std::setw(w) << i << ". " << state.reason->toString(cache, true) << "\n";
+	 if (state.reason->solutions.size() > 1)
+	    out << prefix << std::setw(w) << " " << "  [selected " << it->toString(cache) << " for " << (state.decision == Decision::MUST ? "install" : "remove") << "]\n";
+      }
+      else
+	 out << prefix << std::setw(w) << i << ". " << printSelection(*it, state.decision == Decision::MUST) << "\n";
+   }
+
+   // Print the leaf. We can't have the leaf in the path because we might be called for an attempted decision
+   // that conflicts with the actual assignment (to simplify: we marked X for not install, then we process Y depends X
+   // and try to mark X and reach the conflict, we are called with "X" and the "Y depends X" clause).
+   out << prefix << std::setw(w) << i << ". " << rclause->toString(cache, true) << "\n";
+   if (rclause->solutions.size() > 1)
+      out << prefix << std::setw(w) << " " << "  " << "[selected " << bestReason(rclause, var).toString(cache) << "]\n";
+
+   bool firstContext = true;
+   // Show alternative paths not taken. Consider you have Y depends X|Z; and you mark X
+   // for it; we show above Y -> X as the path, but here we go show why Z was not considered in "Y depends X|Z".
+   for (auto it = path.rbegin(); it != path.rend(); ++it)
+   {
+      auto const &state = (*this)[*it];
+      if (not state.reason) // If we have no reason, we don't have alternatives
+	 continue;
+      if (state.reason->solutions.size() <= 1) // Nothing to print if we no alternatives
+	 continue;
+      if (state.reason->negative) // We only actually need one conflicting choice, ignore others
+	 continue;
+
+      if (firstContext)
+      {
+	 out << prefix << "For context, additional choices that could not be installed:" << "\n";
+      }
+
+      firstContext = false;
+      out << prefix << "* In " << state.reason->toString(cache, true) << ":\n";
+      prefix += "  ";
+      recurseChildren(state.reason, *it);
+      prefix.resize(prefix.size() - 2);
+   }
+   if (rclause->solutions.size() > 1 && not rclause->negative)
+   {
+      if (firstContext)
+      {
+	 out << prefix << "For context, additional choices that could not be installed:" << "\n";
+      }
+      firstContext = false;
+      out << prefix << "* In " << rclause->toString(cache, true) << ":\n";
+      prefix += "  ";
+      recurseChildren(rclause, var);
+      prefix.resize(prefix.size() - 2);
+   }
+
+   return out.str();
+}
+
 // This is essentially asking whether any other binary in the source package has a higher candidate
 // version. This pretends that each package is installed at the same source version as the package
 // under consideration.
@@ -299,7 +501,9 @@ bool APT::Solver::ObsoletedByNewerSourceVersion(pkgCache::VerIterator cand) cons
    for (auto ver = cand.Cache()->FindGrp(cand.SourcePkgName()).VersionsInSource(); not ver.end(); ver = ver.NextInSource())
    {
       // We are only interested in other packages in the same source package; built for the same architecture.
-      if (ver->ParentPkg == cand->ParentPkg || ver.ParentPkg()->Arch != cand.ParentPkg()->Arch || cache.VS->CmpVersion(ver.SourceVerStr(), cand.SourceVerStr()) <= 0)
+      if (ver->ParentPkg == cand->ParentPkg || ver.ParentPkg()->Arch != cand.ParentPkg()->Arch ||
+	    (ver->MultiArch & pkgCache::Version::All) != (cand->MultiArch & pkgCache::Version::All) ||
+	     cache.VS->CmpVersion(ver.SourceVerStr(), cand.SourceVerStr()) <= 0)
 	 continue;
 
       // We also take equal priority here, given that we have a higher version
@@ -316,8 +520,10 @@ bool APT::Solver::ObsoletedByNewerSourceVersion(pkgCache::VerIterator cand) cons
    return false;
 }
 
-bool APT::Solver::Obsolete(pkgCache::PkgIterator pkg) const
+bool APT::Solver::Obsolete(pkgCache::PkgIterator pkg, bool AllowManual) const
 {
+   if ((*this)[pkg].flags.manual && not AllowManual)
+      return false;
    if (pkgObsolete[pkg] != 0)
       return pkgObsolete[pkg] == 2;
 
@@ -336,12 +542,16 @@ bool APT::Solver::Obsolete(pkgCache::PkgIterator pkg) const
    if (ObsoletedByNewerSourceVersion(ver))
       return true;
 
-   for (auto file = ver.FileList(); !file.end(); file++)
-      if ((file.File()->Flags & pkgCache::Flag::NotSource) == 0)
+   // Any version downloadable is good enough for us tbh
+   for (auto ver = pkg.VersionList(); not ver.end(); ++ver)
+   {
+      if (ver.Downloadable())
       {
 	 pkgObsolete[pkg] = 1;
 	 return false;
       }
+   }
+
    if (debug >= 3)
       std::cerr << "Obsolete: " << ver.ParentPkg().FullName() << "=" << ver.VerStr() << " - not installable\n";
    pkgObsolete[pkg] = 2;
@@ -361,7 +571,14 @@ bool APT::Solver::Enqueue(Var var, bool decision, const Clause *reason)
    if (state.decision != Decision::NONE)
    {
       if (state.decision != decisionCast)
-	 return _error->Error("Conflict: %s -> %s%s but %s", WhyStr(bestReason(reason, var)).c_str(), decision ? "" : "not ", var.toString(cache).c_str(), WhyStr(var).c_str());
+      {
+	 std::ostringstream err;
+	 err << "Unable to satisfy dependencies. Reached two conflicting decisions:" << "\n";
+	 std::unordered_set<Var> seen;
+	 err << "1. " << LongWhyStr(var, state.decision == Decision::MUST, state.reason, "   ", seen).substr(3) << "\n";
+	 err << "2. " << LongWhyStr(var, decision, reason, "   ", seen).substr(3);
+	 return _error->Error("%s", err.str().c_str());
+      }
       return true;
    }
 
@@ -445,13 +662,86 @@ bool APT::Solver::Propagate()
    return true;
 }
 
-void APT::Solver::RegisterClause(Clause &&clause)
+static bool SameOrGroup(pkgCache::DepIterator a, pkgCache::DepIterator b)
+{
+   while (1)
+   {
+      if (a->DependencyData != b->DependencyData)
+	 return false;
+
+      // At least one has reached the end, break
+      if (not(a->CompareOp & pkgCache::Dep::Or) || not(b->CompareOp & pkgCache::Dep::Or))
+	 break;
+
+      ++a, ++b;
+   }
+   // Fully iterated over a and b
+   return not(a->CompareOp & pkgCache::Dep::Or) && not(b->CompareOp & pkgCache::Dep::Or);
+}
+
+const APT::Solver::Clause *APT::Solver::RegisterClause(Clause &&clause)
 {
    auto &clauses = (*this)[clause.reason].clauses;
+   pkgCache::DepIterator dep(cache, clause.dep);
+
+   // Merge dependencies on the same name into a single one, and restrict their solution space.
+   // For example, given dependencies on
+   //	 bar Depends: pkg (<< 3), pkg (>> 2)
+   //	 foo Provides: pkg (= 1)
+   // The solution must always be pkg (= 2) and not say pkg (= 3), foo.
+   // FIXME: This would be nice to merge across or groups too, but we can't do that yet.
+   if (not clause.negative && not dep.end() && not(dep->CompareOp & pkgCache::Dep::Or))
+   {
+      bool merged = false;
+      for (auto const &earlierClause : clauses)
+      {
+	 if (earlierClause->negative)
+	    continue;
+	 // Skip dependencies with or groups or dependencies on different names
+	 if (pkgCache::DepIterator earlierDep(cache, earlierClause->dep);
+	     earlierDep.end() || (earlierDep->CompareOp & pkgCache::Dep::Or) ||
+	     earlierDep.TargetPkg() != dep.TargetPkg())
+	    continue;
+	 if (std::none_of(earlierClause->solutions.begin(), earlierClause->solutions.end(), [&clause](auto earlierSol)
+			  { return std::find(clause.solutions.begin(),
+					     clause.solutions.end(),
+					     earlierSol) != clause.solutions.end(); }))
+	    continue;
+
+	 if (earlierClause->optional == clause.optional)
+	 {
+	    std::erase_if(earlierClause->solutions, [&clause, this](auto earlierSol)
+			  { return std::find(clause.solutions.begin(),
+					     clause.solutions.end(),
+					     earlierSol) == clause.solutions.end(); });
+
+	    earlierClause->merged.push_front(clause);
+	    merged = true;
+	 }
+	 else if (clause.optional)
+	 {
+	    // If say a Depends has fewer solution than a Recommends, remove the Recommend's extranous ones.
+	    std::erase_if(clause.solutions, [&earlierClause, this](auto sol)
+			  { return std::find(earlierClause->solutions.begin(),
+					     earlierClause->solutions.end(),
+					     sol) == earlierClause->solutions.end(); });
+
+	    // Remove recursion here, such that we display correctly (if we ever display anywhere...)
+	    auto earlierClauseCopy = *earlierClause;
+	    clause.merged = std::move(earlierClauseCopy.merged);
+	    clause.merged.push_front(earlierClauseCopy);
+	 }
+      }
+
+      if (merged)
+	 return nullptr;
+   }
+
    clauses.push_back(std::make_unique<Clause>(std::move(clause)));
    auto const &inserted = clauses.back();
    for (auto var : inserted->solutions)
       (*this)[var].rclauses.push_back(inserted.get());
+   return inserted.get();
 }
 
 void APT::Solver::Discover(Var var)
@@ -511,8 +801,8 @@ void APT::Solver::Discover(Var var)
 
 	    // This dependency is shared across all versions, skip it.
 	    if (auto &pkgClauses = (*this)[Ver.ParentPkg()].clauses;
-		std::any_of(pkgClauses.begin(), pkgClauses.end(), [start](auto &c)
-			    { return c->dep && c->dep->DependencyData == start->DependencyData; }))
+		std::any_of(pkgClauses.begin(), pkgClauses.end(), [this, start](auto &c)
+			    { return c->dep && SameOrGroup(start, pkgCache::DepIterator(cache, c->dep)); }))
 	       continue;
 
 	    auto clause = TranslateOrGroup(start, end, Var(Ver));
@@ -533,18 +823,21 @@ void APT::Solver::RegisterCommonDependencies(pkgCache::PkgIterator Pkg)
 {
    for (auto dep = Pkg.VersionList().DependsList(); not dep.end();)
    {
-      pkgCache::DepIterator start;
-      pkgCache::DepIterator end;
+      pkgCache::DepIterator start, end;
       dep.GlobOr(start, end); // advances dep
 
       bool allHaveDep = true;
-      for (auto ver = Pkg.VersionList()++; not ver.end(); ver++)
+      for (auto ver = Pkg.VersionList()++; allHaveDep && not ver.end(); ver++)
       {
 	 bool haveDep = false;
-	 for (auto otherDep = ver.DependsList(); not haveDep && not otherDep.end(); otherDep++)
-	    haveDep = otherDep->DependencyData == start->DependencyData;
-	 if (!haveDep)
-	    allHaveDep = haveDep;
+	 for (auto otherDep = ver.DependsList(); not haveDep && not otherDep.end();)
+	 {
+	    pkgCache::DepIterator otherStart, otherEnd;
+	    otherDep.GlobOr(otherStart, otherEnd); // advances other dep
+	    haveDep = SameOrGroup(start, otherStart);
+	 }
+	 if (not haveDep)
+	    allHaveDep = false;
       }
       if (not allHaveDep)
 	 continue;
@@ -555,9 +848,6 @@ void APT::Solver::RegisterCommonDependencies(pkgCache::PkgIterator Pkg)
 
 APT::Solver::Clause APT::Solver::TranslateOrGroup(pkgCache::DepIterator start, pkgCache::DepIterator end, Var reason)
 {
-   auto TgtPkg = start.TargetPkg();
-   auto Ver = start.ParentVer();
-
    // Non-important dependencies can only be installed if they are currently satisfied, see the check further
    // below once we have calculated all possible solutions.
    if (start.ParentPkg()->CurrentVer == 0 && not policy.IsImportantDep(start))
@@ -566,7 +856,7 @@ APT::Solver::Clause APT::Solver::TranslateOrGroup(pkgCache::DepIterator start, p
    if (start->Type == pkgCache::Dep::Replaces || start->Type == pkgCache::Dep::Enhances)
       return Clause{reason, Group::Satisfy, true};
    if (unlikely(debug >= 3))
-      std::cerr << "Found dependency critical " << Ver.ParentPkg().FullName() << "=" << Ver.VerStr() << " -> " << start.TargetPkg().FullName() << "\n";
+      std::cerr << "Found dependency critical " << reason.toString(cache) << " -> " << start.TargetPkg().FullName() << "\n";
 
    Clause clause{reason, Group::Satisfy, not start.IsCritical() /* optional */, start.IsNegative()};
 
@@ -593,8 +883,7 @@ APT::Solver::Clause APT::Solver::TranslateOrGroup(pkgCache::DepIterator start, p
 	    clause.solutions.push_back(Var(pkgCache::VerIterator(cache, *tgt)));
 	 }
 	 delete[] all;
-
-	 std::stable_sort(clause.solutions.begin() + begin, clause.solutions.end(), CompareProviders3{cache, policy, TgtPkg, *this});
+	 std::stable_sort(clause.solutions.begin() + begin, clause.solutions.end(), CompareProviders3{cache, policy, start.TargetPkg(), *this});
       }
       if (start == end)
 	 break;
@@ -616,65 +905,92 @@ APT::Solver::Clause APT::Solver::TranslateOrGroup(pkgCache::DepIterator start, p
 		   { return var.CastPkg(cache)->CurrentVer == 0; }))
       clause.group = Group::SatisfyNew;
    if (std::any_of(clause.solutions.begin(), clause.solutions.end(), [this](auto var) -> auto
-		   { return Obsolete(var.CastPkg(cache)); }))
+		   { return Obsolete(var.CastPkg(cache), true); }))
       clause.group = Group::SatisfyObsolete;
    // Try to perserve satisfied Recommends. FIXME: We should check if the Recommends was there in the installed version?
    if (clause.optional && start.ParentPkg()->CurrentVer)
    {
       bool important = policy.IsImportantDep(start);
-      bool newOptional = true;
-      bool wasImportant = false;
-      for (auto D = start.ParentPkg().CurrentVer().DependsList(); not D.end(); D++)
-	 if (not D.IsCritical() && not D.IsNegative() && D.TargetPkg() == start.TargetPkg())
-	    newOptional = false, wasImportant = policy.IsImportantDep(D);
-
+      auto importantToKeep = [this](pkgCache::DepIterator d)
+      {
+	 return policy.IsImportantDep(d) || (KeepRecommends && d->Type == pkgCache::Dep::Recommends) || (KeepSuggests && d->Type == pkgCache::Dep::Suggests);
+      };
       bool satisfied = std::any_of(clause.solutions.begin(), clause.solutions.end(), [this](auto var)
-				   { return Var(var.CastPkg(cache).CurrentVer()) == var; });
+				   { return var.Pkg(cache) ? var.Pkg(cache)->CurrentVer != nullptr : Var(var.CastPkg(cache).CurrentVer()) == var; });
 
-      if (important && wasImportant && not newOptional && not satisfied)
+      // Find the existing dependency
+      pkgCache::DepIterator existing;
+      for (auto D = start.ParentPkg().CurrentVer().DependsList(); not D.end(); D++)
+	 if (not D.IsCritical() && importantToKeep(D) && D.TargetPkg() == start.TargetPkg())
+	 {
+	    existing = D;
+	    break;
+	 }
+
+      if (not existing.end() && not important && importantToKeep(start) && satisfied)
       {
 	 if (unlikely(debug >= 3))
-	    std::cerr << "Ignoring unsatisfied Recommends " << clause.toString(cache) << std::endl;
-	 clause.solutions.clear();
-      }
-      else if (not important && not wasImportant && not newOptional && satisfied)
-      {
-	 if (unlikely(debug >= 3))
-	    std::cerr << "Promoting satisfied Suggests to Recommends: " << clause.toString(cache) << std::endl;
-	 important = true;
-      }
-      else if (satisfied && important && wasImportant && clause.solutions.size() > 0)
-      {
-	 if (unlikely(debug >= 3))
-	    std::cerr << "Promoting existing Recommends " << clause.toString(cache) << " to depends in upgrade" << std::endl;
-	 clause.optional = false;
-      }
-      else if (newOptional && important && reason.Ver() && clause.solutions.size() > 0 && reason.Ver(cache) != reason.CastPkg(cache).CurrentVer() && IsUpgrade)
-      {
-	 if (unlikely(debug >= 3))
-	    std::cerr << "Promoting new Recommends " << clause.toString(cache) << " to depends in upgrade" << std::endl;
-	 clause.optional = false;
+	    std::cerr << "Try to keep satisfied: " << clause.toString(cache, true) << std::endl;
+	 clause.group = Group::SatisfySuggests;
+	 // Erase the non-installed solutions. We will process this last and try to keep the previously installed
+	 // "best" solution installed.
+	 clause.solutions.erase(std::remove_if(clause.solutions.begin(), clause.solutions.end(), [this](auto var)
+					       { return var.CastPkg(cache)->CurrentVer == nullptr; }),
+				clause.solutions.end());
       }
       else if (not important)
       {
 	 if (unlikely(debug >= 3))
-	    std::cerr << "Ignoring Suggests " << clause.toString(cache) << std::endl;
+	    std::cerr << "Ignore unimportant clause: " << clause.toString(cache, true) << std::endl;
 	 return Clause{reason, Group::Satisfy, true};
       }
+      else if (not existing.end() && policy.IsImportantDep(existing) && not satisfied)
+      {
+	 if (unlikely(debug >= 3))
+	    std::cerr << "Ignoring unsatisfied clause: " << clause.toString(cache, true) << std::endl;
+	 return Clause{reason, Group::Satisfy, true};
+      }
+      else if (IsUpgrade && not existing.end() && satisfied)
+      {
+	 if (unlikely(debug >= 3))
+	    std::cerr << "Promoting previously satisfied clause to hard dependency: " << clause.toString(cache, true) << std::endl;
+	 clause.optional = false;
+	 clause.eager = true;
+      }
+      else if (
+	 IsUpgrade && not(AllowRemove && AllowInstall)				    // promote Recommends to Depends in upgrade, not in dist-upgrade.
+	 && reason.Ver() && reason.Ver(cache) != reason.CastPkg(cache).CurrentVer() // and if this an upgrade to an installed package
+	 && (existing.end() || not policy.IsImportantDep(existing))		    // new Recommends, or upgraded from Suggests
+      )
+      {
+	 if (unlikely(debug >= 3))
+	    std::cerr << "Promoting new clause to hard dependency: " << clause.toString(cache) << std::endl;
+	 clause.optional = false;
+	 clause.eager = true;
+      }
+      else if (not existing.end() && importantToKeep(start) && satisfied)
+      {
+	 if (unlikely(debug >= 3))
+	    std::cerr << "Restricting existing Recommends to installed packages: " << clause.toString(cache, true) << std::endl;
+	 // Erase the non-installed solutions. We will process this last and try to keep the previously installed
+	 // "best" solution installed.
+	 clause.solutions.erase(std::remove_if(clause.solutions.begin(), clause.solutions.end(), [this](auto var)
+					       { return var.CastPkg(cache)->CurrentVer == nullptr; }),
+				clause.solutions.end());
+	 clause.eager = true;
+      }
    }
 
    return clause;
 }
 
-void APT::Solver::Push(Work work)
+void APT::Solver::Push(Var var, Work work)
 {
    if (unlikely(debug >= 2))
       std::cerr << "Trying choice for " << work.toString(cache) << std::endl;
 
    choices.push_back(solved.size());
-   solved.push_back(Solved{Var(), std::move(work)});
-   // Pop() will call MergeWithStack() when reverting to level 0, or RevertToStack after dumping to the debug log.
-   _error->PushToStack();
+   solved.push_back(Solved{var, std::move(work)});
 }
 
 void APT::Solver::UndoOne()
@@ -691,6 +1007,7 @@ void APT::Solver::UndoOne()
       auto &state = (*this)[solvedItem.assigned];
       state.decision = Decision::NONE;
       state.reason = nullptr;
+      state.reasonStr = nullptr;
       state.depth = 0;
    }
 
@@ -713,19 +1030,28 @@ bool APT::Solver::Pop()
    if (depth() == 0)
       return false;
 
+   time_t now = time(nullptr);
+   if (startTime == 0)
+      startTime = now;
+   if (now - startTime >= Timeout)
+      return _error->Error("Solver timed out.");
+
    if (unlikely(debug >= 2))
       for (std::string msg; _error->PopMessage(msg);)
 	 std::cerr << "Branch failed: " << msg << std::endl;
 
-   time_t now = time(nullptr);
-   if (now - startTime >= Timeout)
-      return _error->Error("Solver timed out.");
+   _error->Discard();
 
-   _error->RevertToStack();
+   // Assume() actually failed to enqueue anything, abort here
+   if (choices.back() == solved.size())
+   {
+      choices.pop_back();
+      return true;
+   }
 
    assert(choices.back() < solved.size());
    int itemsToUndo = solved.size() - choices.back();
-   auto choice = solved[choices.back()].work->choice;
+   auto choice = solved[choices.back()].assigned;
 
    for (; itemsToUndo; --itemsToUndo)
       UndoOne();
@@ -743,9 +1069,11 @@ bool APT::Solver::Pop()
       std::cerr << "Backtracking to choice " << choice.toString(cache) << "\n";
 
    // FIXME: There should be a reason!
-   if (not Enqueue(choice, false, {}))
+   if (not choice.empty() && not Enqueue(choice, false, {}))
       return false;
 
+   (*this)[choice].reasonStr = "backtracked";
+
    if (unlikely(debug >= 2))
       std::cerr << "Backtracked to choice " << choice.toString(cache) << "\n";
 
@@ -760,7 +1088,7 @@ bool APT::Solver::AddWork(Work &&w)
 	 if (not Enqueue(var, false, w.clause))
 	    return false;
    }
-   else if (not w.clause->solutions.empty())
+   else
    {
       if (unlikely(debug >= 3 && w.clause->optional))
 	 std::cerr << "Enqueuing Recommends " << w.clause->toString(cache) << std::endl;
@@ -772,15 +1100,13 @@ bool APT::Solver::AddWork(Work &&w)
       work.push_back(std::move(w));
       std::push_heap(work.begin(), work.end());
    }
-   else if (not w.clause->optional && w.clause->dep)
-      return _error->Error("Unsatisfiable dependency group %s -> %s", w.clause->reason.toString(cache).c_str(), pkgCache::DepIterator(cache, w.clause->dep).TargetPkg().FullName().c_str());
-   else if (not w.clause->optional)
-      return _error->Error("Unsatisfiable dependency group %s", w.clause->reason.toString(cache).c_str());
    return true;
 }
 
 bool APT::Solver::Solve()
 {
+   _error->PushToStack();
+   DEFER([&]() { _error->MergeWithStack(); });
    startTime = time(nullptr);
    while (true)
    {
@@ -817,8 +1143,6 @@ bool APT::Solver::Solve()
       if (unlikely(debug >= 1))
 	 std::cerr << item.toString(cache) << std::endl;
 
-      assert(item.clause->solutions.size() > 1 || item.clause->optional);
-
       bool foundSolution = false;
       for (auto &sol : item.clause->solutions)
       {
@@ -830,8 +1154,7 @@ bool APT::Solver::Solve()
 	 }
 	 if (item.size > 1 || item.clause->optional)
 	 {
-	    item.choice = sol;
-	    Push(item);
+	    Push(sol, item);
 	 }
 	 if (unlikely(debug >= 3))
 	    std::cerr << "(try it: " << sol.toString(cache) << ")\n";
@@ -842,15 +1165,13 @@ bool APT::Solver::Solve()
       }
       if (not foundSolution && not item.clause->optional)
       {
-	 std::ostringstream dep;
-	 assert(item.clause->solutions.size() > 0);
-	 for (auto &sol : item.clause->solutions)
-	    dep << (dep.tellp() == 0 ? "" : " | ") << sol.toString(cache);
-	 _error->Error("Unsatisfiable dependency: %s -> %s", WhyStr(item.clause->reason).c_str(), dep.str().c_str());
-	 for (auto &sol : item.clause->solutions)
-	    if ((*this)[sol].decision == Decision::MUSTNOT)
-	       _error->Error("Not considered: %s: %s", sol.toString(cache).c_str(),
-			     WhyStr(sol).c_str());
+	 std::ostringstream err;
+
+	 err << "Unable to satisfy dependencies. Reached two conflicting decisions:" << "\n";
+	 std::unordered_set<Var> seen;
+	 err << "1. " << LongWhyStr(item.clause->reason, true, (*this)[item.clause->reason].reason, "   ", seen).substr(3) << "\n";
+	 err << "2. " << LongWhyStr(item.clause->reason, false, item.clause, "   ", seen).substr(3);
+	 _error->Error("%s", err.str().c_str());
 	 if (not Pop())
 	    return false;
       }
@@ -863,6 +1184,7 @@ bool APT::Solver::Solve()
 bool APT::Solver::FromDepCache(pkgDepCache &depcache)
 {
    DefaultRootSetFunc2 rootSet(&cache);
+   std::vector<Var> manualPackages;
 
    // Enforce strict pinning rules by rejecting all forbidden versions.
    if (StrictPinning)
@@ -878,6 +1200,11 @@ bool APT::Solver::FromDepCache(pkgDepCache &depcache)
       }
    }
 
+   // Clause discovery depends on the manual flag, so we need to set the manual flag first before we discover any packages
+   for (auto P = cache.PkgBegin(); not P.end(); P++)
+      if (P->CurrentVer && not(depcache[P].Flags & pkgCache::Flag::Auto) && (depcache[P].Keep() || depcache[P].Install()))
+	 (*this)[P].flags.manual = true;
+
    for (auto P = cache.PkgBegin(); not P.end(); P++)
    {
       if (P->VersionList == nullptr)
@@ -921,7 +1248,6 @@ bool APT::Solver::FromDepCache(pkgDepCache &depcache)
 	 if (unlikely(debug >= 1))
 	    std::cerr << "Install " << P.FullName() << " (" << (isEssential ? "E" : "") << (isAuto ? "M" : "") << (Root ? "R" : "") << ")"
 		      << "\n";
-
 	 if (not isOptional)
 	 {
 	    // Pre-empt the non-optional requests, as we don't want to queue them, we can just "unit propagate" here.
@@ -932,10 +1258,13 @@ bool APT::Solver::FromDepCache(pkgDepCache &depcache)
 	 {
 	    Clause w{Var(), Group, isOptional};
 	    w.solutions.push_back(Var(P));
-	    RegisterClause(std::move(w));
-	    if (not AddWork(Work{rootState->clauses.back().get(), depth()}))
+	    auto insertedW = RegisterClause(std::move(w));
+	    if (insertedW && not AddWork(Work{insertedW, depth()}))
 	       return false;
 
+	    if (not isAuto)
+	       manualPackages.push_back(Var(P));
+
 	    // Given A->A2|A1, B->B1|B2; Bn->An, if we select `not A1`, we
 	    // should try to install A2 before trying B so we end up with
 	    // A2, B2, instead of removing A1 to keep B1 installed. This
@@ -945,8 +1274,8 @@ bool APT::Solver::FromDepCache(pkgDepCache &depcache)
 	    for (auto V = P.VersionList(); not V.end(); ++V)
 	       shortcircuit.solutions.push_back(Var(V));
 	    std::stable_sort(shortcircuit.solutions.begin(), shortcircuit.solutions.end(), CompareProviders3{cache, policy, P, *this});
-	    RegisterClause(std::move(shortcircuit));
-	    if (not AddWork(Work{rootState->clauses.back().get(), depth()}))
+	    auto insertedShort = RegisterClause(std::move(shortcircuit));
+	    if (insertedShort && not AddWork(Work{insertedShort, depth()}))
 	       return false;
 
 	    // Discovery here is needed so the shortcircuit clause can actually become unit.
@@ -964,17 +1293,29 @@ bool APT::Solver::FromDepCache(pkgDepCache &depcache)
 	 std::stable_sort(w.solutions.begin(), w.solutions.end(), CompareProviders3{cache, policy, P, *this});
 	 if (unlikely(debug >= 1))
 	    std::cerr << "Install essential package " << P << std::endl;
-	 RegisterClause(std::move(w));
-	 if (not AddWork(Work{rootState->clauses.back().get(), depth()}))
+	 auto inserted = RegisterClause(std::move(w));
+	 if (inserted && not AddWork(Work{inserted, depth()}))
 	    return false;
       }
    }
 
-   return Propagate();
+   if (not Propagate())
+      return false;
+
+   std::stable_sort(manualPackages.begin(), manualPackages.end(), CompareProviders3{cache, policy, {}, *this});
+   for (auto assumption : manualPackages)
+   {
+      if (not Assume(assumption, true, {}) || not Propagate())
+	 if (not Pop())
+	    abort();
+   }
+
+   return true;
 }
 
 bool APT::Solver::ToDepCache(pkgDepCache &depcache) const
 {
+   FastContiguousCacheMap<pkgCache::Package, bool> movedManual(cache);
    pkgDepCache::ActionGroup group(depcache);
    for (auto P = cache.PkgBegin(); not P.end(); P++)
    {
@@ -994,10 +1335,31 @@ bool APT::Solver::ToDepCache(pkgDepCache &depcache) const
 
 	 if (cand != P.CurrentVer())
 	 {
+	    bool automatic = (not reason.empty() || (depcache[P].Flags & pkgCache::Flag::Auto)) && not movedManual[P];
 	    depcache.SetCandidateVersion(cand);
-	    depcache.MarkInstall(P, false, 0, reason.empty() && not(depcache[P].Flags & pkgCache::Flag::Auto));
+	    depcache.MarkInstall(P, false, 0, not automatic);
+
+	    // Set the automatic bit for new packages or move it on upgrades to oldlibs
 	    if (not P->CurrentVer)
-	       depcache.MarkAuto(P, not reason.empty());
+	       depcache.MarkAuto(P, automatic);
+	    else if (not(depcache[P].Flags & pkgCache::Flag::Auto) && P.CurrentVer()->Section && cand->Section && not _config->SectionInSubTree("APT::Move-Autobit-Sections", P.CurrentVer().Section()) && _config->SectionInSubTree("APT::Move-Autobit-Sections", cand.Section()))
+	    {
+	       bool moved = false;
+	       for (auto const &clause : (*this)[cand].clauses)
+		  for (auto sol : clause->solutions)
+		  {
+		     // New installs move the auto-bit. TODO: Should we look at whether clause is the reason for installing it?
+		     if (sol.CastPkg(cache) == P || sol.CastPkg(cache)->CurrentVer)
+			continue;
+		     std::cerr << "Move  manual bit from " << P.FullName() << " to " << sol.CastPkg(cache).Name() << std::endl;
+		     movedManual[sol.CastPkg(cache)] = true;
+		     depcache.MarkAuto(sol.CastPkg(cache), false);
+		     moved = true;
+		  }
+	       if (moved)
+		  depcache.MarkAuto(P, true);
+	    }
+
 	 }
 	 else
 	    depcache.MarkKeep(P, false, reason.empty() && not(depcache[P].Flags & pkgCache::Flag::Auto));
@@ -1007,10 +1369,31 @@ bool APT::Solver::ToDepCache(pkgDepCache &depcache) const
       }
       else if (P->CurrentVer || depcache[P].Install())
       {
-	 depcache.MarkDelete(P, false, 0, not(*this)[P].reason);
+	 bool automatic = (not (*this)[P].reason) && (not (*this)[P].reasonStr);
+	 depcache.MarkDelete(P, false, 0, automatic);
 	 depcache[P].Marked = 0;
 	 depcache[P].Garbage = 1;
       }
    }
    return true;
 }
+
+// Command-line
+std::string APT::Solver::InternalCliWhy(pkgDepCache &cache, pkgCache::PkgIterator pkg, bool decision)
+{
+   APT::Solver solver(cache.GetCache(), cache.GetPolicy(), EDSP::Request::Flags(0));
+   // In case nothing has a positive dependency on pkg it may not actually be discovered in a `why-not`
+   // scenario, so make sure we discover it explicitly.
+   solver.Discover(Var(pkg));
+   if (not solver.FromDepCache(cache) || not solver.Solve())
+      return "";
+   std::unordered_set<Var> seen;
+   std::string buf;
+   if (solver[Var(pkg)].decision == Decision::NONE)
+      return strprintf(buf, "%s is undecided\n", pkg.FullName().c_str()), buf;
+   if (decision && solver[Var(pkg)].decision != Decision::MUST)
+      return strprintf(buf, "%s is not actually marked for install\n", pkg.FullName().c_str()), buf;
+   if (not decision && solver[Var(pkg)].decision == Decision::MUST)
+      return strprintf(buf, "%s is actually marked for install\n", pkg.FullName().c_str()), buf;
+   return solver.LongWhyStr(Var(pkg), decision, solver[Var(pkg)].reason, "", seen);
+}
