@@ -11,6 +11,7 @@
 #include <memory>
 #include <optional>
 #include <queue>
+#include <type_traits>
 #include <vector>
 
 #include <apt-pkg/configuration.h>
@@ -201,7 +202,7 @@ class Solver
    // and std::pop_heap() rather than a priority_queue because we need to
    // be able to iterate over the queued work and see if a choice would
    // invalidate any work.
-   heap<Work> work{};
+   heap<Work> work;
 
    // \brief Backlog of solved work.
    //
@@ -209,7 +210,7 @@ class Solver
    // here to revisit it later. This is similar to what MiniSAT calls the
    // trail; one distinction is that we have both literals and our work
    // queue to be concerned about
-   std::vector<Solved> solved{};
+   std::vector<Solved> solved;
 
    // \brief Propagation queue
    std::queue<Var> propQ;
@@ -222,7 +223,7 @@ class Solver
    std::vector<depth_type> choices{};
 
    // \brief The time we called Solve()
-   time_t startTime;
+   time_t startTime{};
 
    EDSP::Request::Flags requestFlags;
    /// Various configuration options
@@ -236,7 +237,7 @@ class Solver
    // \brief If set, removals are allowed.
    bool AllowRemove{_config->FindB("APT::Solver::Remove", not(requestFlags & EDSP::Request::FORBID_REMOVE))};
    // \brief If set, removal of manual packages is allowed.
-   bool AllowRemoveManual{AllowRemove && _config->FindB("APT::Solver::RemoveManual", false)};
+   bool AllowRemoveManual{AllowRemove && _config->FindB("APT::Solver::RemoveManual", true)};
    // \brief If set, installs are allowed.
    bool AllowInstall{_config->FindB("APT::Solver::Install", not(requestFlags & EDSP::Request::FORBID_NEW_INSTALL))};
    // \brief If set, we use strict pinning.
@@ -277,7 +278,7 @@ class Solver
 
    public:
    // \brief Create a new decision level.
-   void Push(Work work);
+   void Push(Var var, Work work);
    // \brief Revert to the previous decision level.
    [[nodiscard]] bool Pop();
    // \brief Undo a single assignment / solved work item
@@ -287,6 +288,7 @@ class Solver
 
    // \brief Basic solver initializer. This cannot fail.
    Solver(pkgCache &Cache, pkgDepCache::Policy &Policy, EDSP::Request::Flags requestFlags);
+   ~Solver();
 
    // Assume that the variable is decided as specified.
    [[nodiscard]] bool Assume(Var var, bool decision, const Clause *reason = nullptr);
@@ -316,6 +318,9 @@ class Solver
     * \param seen A set of seen objects such that the output does not repeat itself (not for safety, it is acyclic)
     */
    std::string LongWhyStr(Var var, bool decision, const Clause *rclause, std::string prefix, std::unordered_set<Var> &seen) const;
+
+   // \brief Temporary internal API with external linkage for the `apt why` and `apt why-not` commands.
+   APT_PUBLIC static std::string InternalCliWhy(pkgDepCache &depcache, pkgCache::PkgIterator Pkg, bool decision);
 };
 
 }; // namespace APT
@@ -401,9 +406,15 @@ struct APT::Solver::Clause
    // \brief A negative clause negates the solutions, that is X->A|B you get X->!(A|B), aka X->!A&!B
    bool negative;
 
-   inline Clause(Var reason, Group group, bool optional = false, bool negative = false) : reason(reason), group(group), optional(optional), negative(negative) {}
+   // \brief An optional clause may be eager
+   bool eager;
 
-   std::string toString(pkgCache &cache, bool pretty = false) const;
+   // Clauses merged with this clause
+   std::forward_list<Clause> merged;
+
+   inline Clause(Var reason, Group group, bool optional = false, bool negative = false) : reason(reason), group(group), optional(optional), negative(negative), eager(not optional) {}
+
+   std::string toString(pkgCache &cache, bool pretty = false, bool showMerged = true) const;
 };
 
 /**
@@ -423,15 +434,8 @@ struct APT::Solver::Work
    // \brief The depth at which the item has been added
    depth_type depth;
 
-   // This is a union because we only need to store the choice we made when adding
-   // to the choice vector, and we don't need the size of valid choices in there.
-   union
-   {
-      // The choice we took
-      Var choice;
-      // Number of valid choices
-      size_t size{0};
-   };
+   // Number of valid choices
+   size_t size{0};
 
    // \brief This item should be removed from the queue.
    bool erased{false};
@@ -473,6 +477,8 @@ struct APT::Solver::State
    // Vars < 0 are package ID, reasons > 0 are version IDs.
    const Clause *reason{};
 
+   const char *reasonStr{};
+
    // \brief The depth at which the decision has been taken
    depth_type depth{0};
 
