@@ -10,7 +10,7 @@
 #include <string>
 #include <vector>
 
-#include <gtest/gtest.h>
+#include "common.h"
 
 #include "file-helpers.h"
 
